'
' -----[ Program Description ]--------------------------------------------------
'
' This program implements the ModBus RTU protocol in an universal I/O device
' It uses a special designed HW or STK200 development kit for testing purpouse.
' It's a polled slave with 16 in and 16 outputs.
'
' -----[ Disclaimer ]-----------------------------------------------------------
'
' This example is offered on an "AS IS" basis, no warranty expressed or implied.
' The programers disclaim liability of any damages associated with the use of
' the hardware or software described herein.
' You use it on your own risk. We are not able to provide any free support.
'
' Copyright (c) 2001 Mike Eitel all rights reserved
'
' -----[ Revision History ]-----------------------------------------------------
'
' 011121 - Version 1.00AVR            Ver 1.00 slave AVR >8515        Mike Eitel
' 011128 - Version 1.10AVR            Ver 1.10 slave AVR  2313        Mike Eitel
' 011221 - Version AVRMODDIO16        Ver 1.20 slave AVR 2313 ERAM    Mike Eitel
' 020914 - Version AVRMODDIO16        Ver 1.30 slave AVR 2313 ERAM    Mike Eitel
' 030410 - Version AVRMODDIO16        Ver 1.50 slave mclib 2313 ERAM  Mike Eitel
'
' -----[ Basic HW definitions ]-------------------------------------------------

' *****[ conditional compiler flags ]*******************************************
Const Test = 1                               ' When 1 then Test !
'#if Test                                     ' ***** TEST MODE ***
'#else                                        ' ***** TEST MODE ***
'#endif                                       ' ***** TEST MODE ***

Const Stk200 = 1                             ' When 1 then STK200 else DTR103
'#if Stk200                                   ' ***** STK200 device ***
'#else                                        ' ***** STK200 device ***
'#endif                                       ' ***** STK200 device ***
' *****[ conditional TEST flag ]************************************************


$regfile = "2313def.dat"
$lib "modbus.lbx"                            ' Modbus.lib contains crcMB func.

' Define serial communication parameters
'Ucr = &B11011000                             ' Allow receiver & transmitter ISR
$baud = 9600                                 ' ModBus speed

#if Stk200                                   ' ***** TEST MODE ***
    $crystal = 4000000                       ' Frequency of unit
    'Ubrr = 51                                ' For 4800 baud
    Ubrr = 25                                ' For 9600 baud
#else                                        ' ***** TEST MODE ***
'    $crystal = 4433618                       ' Frequency of unit
'    Ubrr = 57                                ' For 4800 baud
'    Ubrr = 28                                ' For 9600 baud

'    $crystal = 6144000                       ' Frequency of unit
'    Ubrr = 79                                ' For 4800 baud

    $crystal = 8000000                       ' Frequency of unit
'    Ubrr = 103                               ' For 4800 baud
    Ubrr = 51                                ' For 9600 baud

'    $crystal = 9830488                       ' Frequency of DTR103
'    Ubrr = 63                                ' For 9600 baud
#endif
' -----[ Aliases ]--------------------------------------------------------------


' -----[ Constants ]------------------------------------------------------------

'                            Telegramms solved
'         This device is capable of max 4 bytes data transfer mode 03 & 16 !!!
'
' Request Read    Holding Register(03)  SS 03 SH SL PH PL CL CH
' Answer  Read 16 Holding Register(03)  SS 03 BC 1H 1L CL CH
' Answer  Read 32 Holding Register(03)  SS 03 BC 1H 1L 2H 2L CL CH
'                                                       
'                                   DB( 1  2  3  4  5  6  7  8  9 )
'
' Request Preset Single Register(06)    SS 06 RH RL 1H 1L CL CH
' Answer  Preset Single Register(06)    SS 06 RH RL 1H 1L CL CH
'                                                     
'                                   DB( 1  2  3  4  5  6  7  8 )
'
' Request Preset Multiple Register(16)  SS 10 SH SL PH PL BC 1H 1L 2H 2L CL CH
' Answer  Preset Multiple Register(16)  SS 10 SH SL PH PL CL CH
'                                                               
'                                   DB( 1  2  3  4  5  6  7  8  9  10 11 12 13 )
'
'                                            ' SS = Slave Address
'                                            ' 03 = Command 03
'                                            ' SH = Starting Address High
'                                            ' SL= Starting Address Low
'                                            ' PH = Nr. of Points High
'                                            ' PL = Points Nr.(16 bits) Low
'                                            ' CL = CRC Low
'                                            ' CH = CRC High
'                                            ' BC = Byte Count
'                                            ' 1H = Data 1 High
'                                            ' 1L = Data 1 Low
'                                            ' 2H = Data 2 High
'                                            ' 2L = Data 2 Low
'

Const Initializing = &HFE                    ' Address to change Myaddress
Const Adressparamter = 10                    ' Slaveparameters position
Const Parameters = 9                         ' Slaveparameters position

Const Dataarea = 16                          ' Databuffer + 2 Slaveparameters
Const Savetoeram = 10                        ' Address of Myaddress

Const Maxchar = 19                           ' Nr. of characters for serial buffer
Const Telegrambytes = 14                     ' Nr. of transmitted Bytes
Const Pause = 2                              ' Wait time befor answer

' Pseudo constants programmed via telegram
' Const Myaddress = &H02                       ' Address for this node (1-255)
Dim Myaddress As Eram Byte At &H02           ' Address for this node (2-253)
Dim Myparameter As Eram Byte At &H03         ' Address for this node (2-253)


' -----[ Variables ]------------------------------------------------------------

Dim Datas(dataarea) As Byte At &H60          ' Place to keep the dynamic datas
Dim Bytebuffer(maxchar) As Byte              ' Receive serial-buffer
Dim Db(telegrambytes) As Byte                ' Datas in telegramm
Dim Sdb(telegrambytes) As Byte               ' Datas in send telegramm
Dim Sendbytes As Byte                        ' Number of transmitted Bytes
Dim Sdbcounter As Byte                       ' Lenght of send telegram
Dim Sdbcountertmp As Byte                    ' Helper for send telegram
Dim Datalength As Byte                       ' Recognize the requested commands

Dim Bytesreceived As Byte                    ' Signalling several received characters
Dim Bytecounter As Byte                      ' Counter of received characters
Dim Bytereader As Byte                       ' Counter of transformed characters
Dim Crc As Word                              ' CRC word in communication

Dim Lastbyte As Byte                         ' Last Lowbyte of the inputs

Dim Multiplexa As Byte                       ' Multiplexcounter 16In- & 16Outputs
'Dim Multiplexb As Byte                       ' Helper for multiplexer
Dim Multiplexc As Byte                       ' Helper pointer for fast I/O reaction
Dim Multiplexd As Byte                       ' Helper for fast I/O reaction

Dim I As Byte                                ' Helper for looping
Dim K As Byte                                ' Helper for looping
Dim L As Byte                                ' Helper for looping
Dim M As Byte                                ' Helper for looping

'Dim Countertest As Long At &H60 Overlay
' -----[ Initialization ]---------------------------------------------
'
' Define port parameters
' Port B Is Used For Data Exchange With The Latches
  'Ddrb = &HFF                               ' WRITE to multiple I/O Data port
  'Ddrb = &H00                               ' READ from multiple I/O Data port

'  Port D  is used for communication and for latch controll
  Ddrd = &B01111100                          ' Set Portd Pin 2..6 as Outputs
'  Ddrd.0                          ' RXD
'  Ddrd.1                          ' TXD
'  Ddrd.2                          ' Strobe Enable Read chip 1
'  Ddrd.3                          ' Strobe Enable Read chip 2
'  Ddrd.4                          ' Strobe output latches Allegro UCN 5801A  1
'  Ddrd.5                          ' Strobe output latches Allegro UCN 5801A  1
'  Ddrd.6                          ' Enable for RS422 send pin
'  Ddrd.7                          ' NC for 2313
  Portd = &B00001100

  ' Setup interrupts to defined startup
  Ucr = &B11011000                           ' Allow receive & transmit ISR
  ' Enable Utxc                               ' Disable send isr
  ' Enable Urxc                               ' Enable receive isr
  ' Disable Udre                              ' Enable empty send buffer isr

  On Urxc Rec_isr                            ' Define serial receive ISR
  On Utxc Send_isr                           ' Define serial send ISR
  On Udre Ude_isr Nosave                     ' Define empty serial buffer ISR
  Enable Serial                              ' Allow protokoll treatment starts

  Config Timer0 = Timer , Prescale = 64      ' Define update speed
  On Timer0 Timer0_isr                       ' Define ISR when timer 0 overflows                                         '
  Enable Timer0                              ' Allow I/O treating

  ' Transfer from ERAM slave address and parameters
  Datas(adressparamter) = Myaddress          ' Set new Slave address
  Datas(parameters) = Myparameter            ' Set new Slave parameters

#if Test                                     ' ***** TEST MODE ***
  Datas(adressparamter) = 2                  ' Set test address
  Datas(parameters) = 1                      ' Set test parameters
#endif                                       ' ***** TEST MODE ***

  Enable Interrupts                          ' Start now the system timing

'------[ Program ]----------------------------------------------------
'
_com_z:
   ' Startingpoint for telegrams
   If Bytesreceived >= 1 Then                ' Received something
      Bytereader = Bytecounter - Bytesreceived
      Incr Bytereader
      Decr Bytesreceived
      Db(1) = Bytebuffer(bytereader)         ' transfer to receive telegram buffer

      ' If received address is this node read next Byte from master,
      ' if not return to start
      If Db(1) <> Datas(adressparamter) Then ' Normal address
         If Db(1) <> Initializing Then       ' Address to change Node Nr.
            Goto _com_z                      ' Wait for next incoming data
         End If
      End If
   Else
      Goto _com_z                            ' Wait for next incoming data
   End If

_com_1:
   If Bytesreceived >= 1 Then                ' Received something
      Incr Bytereader
      Decr Bytesreceived
      If Bytereader >= Maxchar Then          ' Buffer full=generate unread message
         Bytereader = 1                      ' Rewrap character serial counter
      End If
      Datalength = 6
      Db(2) = Bytebuffer(bytereader)         ' transfer into receive telegram buffer
      Select Case Db(2)
          Case 3 :                           ' A read command
             Goto _com_2:                    ' Read next values
          Case 6 :                           ' A read command
             Goto _com_2:                    ' Read next values
          Case 16 :                          ' A write command
             Datalength = 11
             Goto _com_2:                    ' Read next values
      End Select
      Goto _com_z                            ' Wrong command = restart
   Else
      Goto _com_1                            ' Wait for next incoming data
   End If

_com_2:
   ' Get telegram into binary mode
   If Bytesreceived < Datalength Then
      Goto _com_2                            ' Wait for next incoming data
   End If
   Incr Datalength                           ' Faster way to add 2
   Incr Datalength                           ' Faster way to add 2

   For I = 3 To Datalength                   ' Transfer rest of telegram data
     Incr Bytereader
     If Bytereader >= Maxchar Then           ' Buffer full=generate unread message
        Bytereader = 1                       ' Rewrap character serial counter
     End If
     Db(i) = Bytebuffer(bytereader)          ' Data(x) transfer to memory
   Next I

   ' Clear the receive buffer to make a calculated response possible
   Bytecounter = 1                           ' Reset for next telegram
   Bytesreceived = 0                         ' Reset for next telegram

   ' Check the CRC of the message for errors

   Decr Datalength                           ' Faster way to subtract 2
   Decr Datalength                           ' Faster way to subtract 2
   Crc = Crcmb(db(1) , Datalength)           ' Use Modbus CRC routine

   If Bytebuffer(bytereader) = High(crc)then ' First check last crc byte
      If Bytebuffer(bytereader - 1) = Low(crc)then       ' Then the lower byte
         Goto _com_4                         ' CRC was ok
      End If                                 ' Lower CRC
   End If                                    ' Higher CRC
   Goto _com_z                               ' Restart undone

_com_4:
   ' No CRC errors in packet so check what to do and start reading / writing
   For I = 1 To 6                            ' Max lenth of ack telegram
       Sdb(i) = Db(i)                        ' Copie for answer
   Next I

   K = Db(4)                                 ' Data start register
   If K > Dataarea Then                      ' Protection agianst too big param.
      Goto _com_z
   End If
   Incr K                                    ' Modbus addaption ..1 = first
   Select Case Db(2)                         ' Decide read or write

      Case 03 :                              ' READING datas 1..n
         ' Request Read    Holding Register(03)   SS 03 SH SL PH PL CL CH
         ' Answer  Read 16 Holding Register(03)   SS 03 BC 1H 1L CL CH
         ' Answer  Read 32 Holding Register(03)   SS 03 BC 1H 1L 2H 2L x  x CL CH
         '                                                         .  .
         '                                     DB(1  2  3  4  5  6  7  8  9 x  x

         Sendbytes = 3                       ' Initial value send corpus
         Sdb(3) = Db(6)                      ' Answer with x words for BC
         M = 4                               ' Telegram Db(x)start byte

         L = 1                               ' Number of start word
         While L <= Sdb(3)                   ' For Nr of requested words
             Sdb(m) = Datas(k + 1)           ' Reading first
             Sdb(m + 1) = Datas(k )          ' Reading first
             Incr Sendbytes                  ' Corpus + accumulated 2 databytes
             Incr Sendbytes                  ' Corpus + accumulated 2 databytes
             Incr K                          ' Next Modbus
             Incr K                          ' Next Modbus
             Incr M                          ' Next node byte
             Incr M                          ' Next node byte
             Incr L                          ' Next node byte
         Wend
         Shift Sdb(3) , Left , 1             ' Answer transformed to bytes


         Gosub _send                         ' Give answer
         Goto _com_z                         ' All done, go back to Start

      Case 06 :                              ' WRITING datas 1..n
         ' Request Preset Single Register(06)      SS 06 RH RL 1H 1L CL CH
         '  Answer Preset Single Register(06)      SS 06 RH RL 1H 1L CL CH
         '                                                       
         '                                     DB( 1  2  3  4  5  6  7  8 )
         '
         Sendbytes = 6                       ' ST,COmmand  plus 6 nibbles
         Datas(k) = Db(6)                    ' Reading first
         Incr K                              ' Next byte
         Datas(k) = Db(5)                    ' Reading second
         If Db(1) = Initializing Then        ' Test for general access
            If K = Dataarea Then             ' Set Parameters to EERAM
               Myparameter = Db(6)           ' Set new Slave parameters
               If Db(5) < 253 Then           ' Usefull address < 253
                  If Db(5) > 1 Then          ' Address starts at 2
                     Myaddress = Db(5)       ' Set new Slave address
                  End If                     ' End address test high
               End If                        ' End address test low
            End If                           ' End Test parameters setting
         End If                              ' End initialisating address
         Gosub _send                         ' Give answer
         Goto _com_z                         ' All done, go back to Start

      Case 16 :                              ' WRITING
  ' Request Preset Multiple Register(16)  SS 10 SH SL PH PL BC 1H 1L 2H 2L CL CH
  '  Answer Preset Multiple Register(16)  SS 10 SH SL PH PL CL CH
  '                                                               
  '                                   DB( 1  2  3  4  5  6  7  8  9  10 11 12 13)
         Sendbytes = 6                       ' Initial value send corpus
         M = 8                               ' Telegram data Db(x)start byte
         '
         For L = 1 To Db(6)                  ' For Nr of requested words
             Datas(k) = Db(m + 1)            ' Reading first
             Datas(k + 1) = Db(m )           ' Reading first
             Incr K                          ' Next Modbus
             Incr M                          ' Next node byte
             Incr K                          ' Next Modbus
             Incr M                          ' Next node byte
         Next L
         Gosub _send                         ' Give answer
         Goto _com_z                         ' All done, go back to Start
      End Select
Goto _com_z                                  ' Not identified command
End                                          ' End of the main program
' -----[ End ]--------------------------------------------------------



' -----[ Subroutines ]------------------------------------------------

' Subroutine for sending values
_send:
   ' Start CRC generate routine and calculate CRC for all sending bytes
   Crc = Crcmb(sdb(1) , Sendbytes )          ' Use Modbus CRC routine
   Incr Sendbytes                            ' Take the next
   Sdb(sendbytes ) = Low(crc)                ' First the low byte
   Incr Sendbytes                            ' take the next
   Sdb(sendbytes ) = High(crc)               ' First the high byte
   Incr Sendbytes                            ' Increment for send buffer size
_again:
   'Send Packet To Master , Including The Sync Byte
   If Sdbcounter < 1 Then                    ' Safety loop, case of to much sending
      Set Portd.6                            ' Set switch to send modus pin
      Sdbcounter = Sendbytes - 2             ' Set buffer to full lenght
      Waitms Pause                           ' Switch over time
      Udr = Sdb(1)                           ' Start ISR work by 1.Byte sending
   Else
      Goto _again                            ' Safety loop should never run
   End If
Return


' -----[ Interrupt routines ]------------------------------------------------

' Interrupt routine for preparing serial port
' On powerup it's treated once !
' This routine should not be needed but only with this routine AVR works fine !
Ude_isr:
   Ucr = &B11011000                          ' Number of character flag
Return


' Interrupt routine for reading serial input
Rec_isr:
   If Bytecounter < Maxchar Then             ' Does it fit into the buffer?
      Incr Bytecounter                       ' Increase buffer counter
   Else
      Bytecounter = 1                        ' Safety rewrap character serial counter
   End If
   Bytebuffer(bytecounter) = Udr             ' Add to buffer array
   Incr Bytesreceived
Return


' Interrupt routine for sendig serial output
Send_isr:
   If Sdbcounter > 0 Then                    ' (0) = Dirty trick to see buffer size ?
      Sdbcountertmp = Sendbytes - Sdbcounter
      Udr = Sdb(sdbcountertmp)               ' Increase buffer counter
      Decr Sdbcounter
   Else
      Reset Portd.6                          ' Clean switch to send modus pin
   End If
Return


' Interrupt routine for multiplexing In / Outputs
Timer0_isr:                                  ' Controlls the multiplexing-----
'                                                    ---- of data versus I/O's
'   Port D is used for communication and for latch controll
'   Ddrd.2                          ' Strobe Enable Read chip 1
'   Ddrd.3                          ' Strobe Enable Read chip 2
'   Ddrd.4                          ' Strobe output latches Allegro UCN 5801A  1
'   Ddrd.5                          ' Strobe output latches Allegro UCN 5801A  2
'   Ddrd.6                          ' Enable for RS422 send pin

    Select Case Multiplexa

    Case 0:                                  ' READ 1 --- High Byte ---
            Reset Portd.2                    ' Set gate of first reading chip
         #if Test                            '                 ***** TEST MODE *
            Ddrb = &HFF                      ' Switch WRITE multiplex I/O Data
            Portb = Datas(1)                 ' Write data to multipexer I/O
         #else                               '                 ***** TEST MODE *
            Ddrb = &H00                      ' Switch READ multiplex I/O Data
            Datas(1) = Pinb                  ' Read data from multipexer I/O
         #endif                              '                 ***** TEST MODE *

    Case 1:                                  ' READ 2 --- Low Byte ---
            Reset Portd.3                    ' Set gate of second reading chip
         #if Test                            '                 ***** TEST MODE *
            Ddrb = &HFF                      ' Switch WRITE multiplex I/O Data
            Portb = Datas(2)                 ' Write data to multipexer I/O
         #else                               '                 ***** TEST MODE *
            Ddrb = &H00                      ' Switch READ multiplex I/O Data
            Datas(2) = Pinb                  ' Read data from multipexer I/O
         #endif                              '                 ***** TEST MODE *
            Multiplexc = Datas(1) And Datas(parameters)       ' Check only
            If Lastbyte <> Multiplexc Then   '                  wished pins
               Lastbyte = Multiplexc         ' Store modified value
               Incr Datas(7)                 ' Counter low of checked pins
               If Datas(7) = 0 Then
                  Incr Datas(8)              ' Counter high of checked pins
               End If                        ' End second byte
            End If                           ' End counting

    Case 2:                                  ' WRITE 3 --- High Byte ---
            If Datas(adressparamter) = &HFF Then       ' Unprogrammed =Test outs
               Datas(3) = Datas(1)           ' Increment the one byte
               Datas(4) = Datas(2)           ' Decrement the other byte
            End If
            Ddrb = &HFF                      ' Switch WRITE multiplex I/O Data
            ' If db(5) is requesting high input pin, force output high. When
            ' input reaches high level forcing is surpressed. Generally output
            ' is OR overwriten high by db(3)
            Multiplexc = Not Datas(1)        ' Check input pins for high status
            Multiplexd = Datas(5) And Multiplexc       ' Compare demand and
            Portb = Multiplexd Or Datas(3)   ' set. Transfer the data to I/O
            Set Portd.4                      ' Set strobe for first writing chip

    Case 3:                                  ' WRITE 4 --- Low Byte ---
            Ddrb = &HFF                      ' Switch WRITE multiplex I/O Data
            ' If db(6) is requesting high input pin, force output high. When
            ' input reaches high level forcing is surpressed. Generally output
            ' is OR overwriten high by db(4)
            Multiplexc = Not Datas(2)        ' Check input pins for high status
            Multiplexd = Datas(6) And Multiplexc       ' Compare demand and
            Portb = Multiplexd Or Datas(4)   ' set. Transfer the data to I/O
            Set Portd.5                      ' Set strobe for second writing chip
    End Select

    Incr Multiplexa                          ' Cycle counter for the datas 1..4
    Multiplexa = Multiplexa And &H03         ' To cut down cycle counter to 0 ..3

'    Incr Countertest
    Set Portd.2                              ' Set bit for first reading chip
    Set Portd.3                              ' Set bit for second reading chip
    Reset Portd.4                            ' Set bit for first writing chip
    Reset Portd.5                            ' Set bit for second writing chip
Return